import os
import click
from jais.utils import load_config, ROOT_DIR, get_logger
from pyfiglet import Figlet

# Load project info
CONTEXT_SETTINGS = dict(auto_envvar_prefix='COMPLEX')
# Load Config file
CNF = load_config(os.getenv('JAIS_CNF_PATH'))


# ----------------------------------------> LOGGING :
# create logger
LOG = get_logger(CNF)
# #! Example logging
# LOG.debug('debug message')
# LOG.info('info message')
# LOG.warning('warn message')
# LOG.error('error message')
# LOG.critical('critical message')


# ----------------------------------------> CLI :
class ComplexCLI(click.MultiCommand):
    """CLI modules finder and loader"""

    def list_commands(self, ctx) -> list:
        """
        This function creates a list of command (`cli.py`) files

        Args:
            ctx: context manager.

        Returns:
            list of CLI files names.
        """
        cli_files_list = []
        # Look for modules that should be loaded with CLI
        for name in ROOT_DIR.iterdir():
            # Remove any __pycache__ like folders
            if str(name).startswith("__"):
                continue
            # Get full path of the module
            dirname = ROOT_DIR/name
            # Look for dtip and dkip like folders
            if dirname.is_dir():
                # name of the folder that contains `cli.py` file
                cli_files_list.append(name)
        cli_files_list.sort()
        return cli_files_list

    def get_command(self, ctx, name: str) -> click.core.Command:
        """This function imports commands from `cli.py` files

        Args:
            ctx: click context object
            name: name of the folder from which cli.py file should be loaded

        Returns:
            returns commands.
        """

        try:
            # {name} is the package like `dtip` and `dkip` and
            # `.cli` is the `cli.py` file inside these.
            mod = __import__(f"jais.{name}.cli", None, None, ["cli"])
        except ImportError:
            return

        return mod.cli


# * Entry point
@click.group(cls=ComplexCLI, invoke_without_command=True, context_settings=CONTEXT_SETTINGS)
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        # * BANNER
        # Find more fonts here: http://www.figlet.org/examples.html
        f = Figlet(font='smslant')
        banner = 'J A I S'
        # banner = f"..._ {banner} _..."
        click.secho(f"{f.renderText(banner)}", fg='yellow')
        click.echo(f"""Welcome to {click.style('Just Artificial Intelligence Snippets (JAIS)', fg='yellow')} CLI.

    Type `{click.style(f"jais --help", fg='yellow')}` for usage details
    """)
        click.echo(ctx.get_help())

    else:
        click.secho(f"\n[@ {ctx.invoked_subcommand}]...", fg='cyan')

if __name__ == '__main__':
    cli()



import pytest
from jais.utils import load_config, ROOT_DIR

# * this function loads the config file for the tests.
@pytest.fixture
def CNF():
    return load_config(ROOT_DIR/"configs/default.yaml")

def test_config(CNF):
    """Check if the config file is loaded correctly."""
    # Check project name
    assert CNF.project_name == "jais"

def test_logging(CNF):
    """Check logging settings"""
    import configparser
    logs_conf = configparser.RawConfigParser()
    logs_conf.read(ROOT_DIR/"configs/logs.conf")
    loggers = logs_conf['loggers']['keys'].split(',')
    assert (CNF.project_name in loggers[-1].lower()) == True


import os
import torch
import torch.nn as nn
from typing import Tuple, Optional
from jais.train.trainer import Trainer
from jais.utils import get_logger, load_config

CNF = load_config(os.getenv('JAIS_CNF_PATH'))
LOG = get_logger(CNF, os.getenv('JAIS_LOG_PATH'))

class TuneHP:
    def __init__(self, dataloader, net, loss_fn, optm_fn, device) -> None:
        self.dataloader = dataloader
        self.net = net.to(device)
        self.loss_fn = loss_fn
        self.optm_fn = optm_fn
        self.device = device

    def tune(self, 
             lr_range: Tuple[float, float] = (1., 1e-6),
             wd_range: Tuple[float, float] = (1e-3, 1e-8),
             mom_range: Tuple[float, float] = (0.9, 0.99),
             optm_fn: Optional[nn.Module] = None,
             max_iters: int = 100):

        # set new optimizer, if given
        if optm_fn:
            self.optm_fn = optm_fn
            LOG.info(f"Optimizer changed to {optm_fn}")
        # Find best learning rate range
        # Train on one batch and validate on next
        #TODO ADD LEARNING RATE FINDER CODE HERE
        dl = iter(self.dataloader)
        for i in range(0, max_iters, 2):
            # Train on one batch
            self.net.train()
            self.train_one_batch(next(dl))
            # Validate on next batch
            self.net.eval()
            with torch.no_grad():
                self.val_one_batch(next(dl))
        

    def train_one_batch(self, batch):
        return Trainer.train_one_batch(batch)
    
    def val_one_batch(self, batch):
        return Trainer.val_one_batch(batch)



import os
from pathlib import Path
from typing import Dict, Union
from torch import Tensor
from numpy import ndarray
from pandas import DataFrame
from rich import print
from rich.live import Live
from rich.progress import Progress, SpinnerColumn, BarColumn, TextColumn
from rich.table import Table
from jais.utils import load_default_configs


__all__ = ['TrainingLogger', 'RichProgress']


CNF, LOG = load_default_configs()


class TrainingLogger:
    def __init__(self, 
                 logger_name: str = None, 
                 log_dir: Union[Path, str] = CNF.paths.logs_dir) -> None:
        """
        Metrics logs handling class for training and validation per 
            epoch or batch

        Args:
            logger_name: name of the logger to use. 
                [wandb (wb), tensorboard (tb), locallogger (ll)]
        """
        self.logger_name = logger_name
        self.log_dir = log_dir
        if self.logger_name:
            if self.logger_name in ['wb', 'wandb']:
                import wandb
                wandb.init(project=CNF.project_name, 
                        dir=self.log_dir, 
                        save_code=True)
                self.logger = wandb.log
            elif self.logger_name in ['tb', 'tensorboard']:
                from torch.utils.tensorboard import SummaryWriter

                self.writer = SummaryWriter(log_dir=f"{self.log_dir}/tensorboard")
            else:
                raise ValueError("Only `wandb` and `tensorboard` is supported.")

    def log(self, logs: Dict[str, Union[Tensor, ndarray, float]]) -> None:
        """Add the metrics to the metrics list

        Args:
            logs: dict of values to add in the history
        """        
        # Log to the particular logger
        if self.logger_name:
            if self.logger_name in ['tb', 'tensorboard']:
                for k, v in logs.items():
                    if k.endswith('_loss'):
                        self.writer.add_scalar(
                            f"Loss/{k.split('_')[0]}", 
                            v, 
                            global_step=logs['epoch_num'])
                    else:
                        self.writer.add_scalar(k, v, global_step=logs['epoch_num'])
            elif self.logger_name in ['wb', 'wandb']:
                self.logger(logs)
            else:
                _errmsg = "Only `wandb` and `tensorboard` is supported."
                raise NotImplementedError(_errmsg)


class RichProgress:
    def __init__(self, 
                 columns: list, 
                 n_epochs: int, 
                 n_train_iters: int, 
                 n_val_iters: int, 
                 refresh_rate: int = 10) -> None:
        self.columns, self.n_epochs, self.epoch_num = columns, n_epochs, 1
        self.n_train_iters, self.n_val_iters = n_train_iters, n_val_iters

        self.progress_bar = Progress(
            "{task.description}",
            SpinnerColumn(),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            transient=True
        )
        self.epoch_bar = self.progress_bar.add_task(
            f"[cyan]Overall", total=self.n_epochs
        )
        self.train_bar = self.progress_bar.add_task(
            f"Epoch [1/{self.n_epochs}]", 
            total=self.n_train_iters, 
            visible=False
        )
        self.val_bar = self.progress_bar.add_task(
            f"Validate", total=self.n_val_iters, visible=False
        )

        # Make Table and Bars
        progress_table = Table.grid(expand=True)
        self.metrics_table = Table()
        for col in self.columns:
            self.metrics_table.add_column(col, justify='center')
        progress_table.add_row(self.metrics_table)
        progress_table.add_row(self.progress_bar)

        self.live = Live(progress_table, 
                         refresh_per_second=refresh_rate, 
                         vertical_overflow=None)
                         
    def __enter__(self) -> None:
        self.live.start()
        return self

    def update_epoch(self, row: list) -> None:
        """Update metrics table and progress"""

        typecast = lambda x: f"{x:.4f}" if isinstance(x, float) else str(x)

        self.reset_progress_bars()
        self.metrics_table.add_row(*map(typecast, row))
        self.progress_bar.update(self.epoch_bar, advance=1)
        self.epoch_num += 1 # For self.train_bar value
        
    def update_train_bar(self) -> None:
        """Update the training progress bar visualizations"""
        description = f"Epoch [{self.epoch_num}/{self.n_epochs}]"
        self.progress_bar.update(self.train_bar,
                                 description=description,
                                 advance=1, visible=True)
        self.progress_bar.update(self.val_bar, visible=False)
        
    def update_val_bar(self) -> None:
        """Update the validation progress bar visualizations"""
        self.progress_bar.update(self.train_bar, visible=False)
        self.progress_bar.update(self.val_bar, advance=1, visible=True)

    def reset_progress_bars(self) -> None:
        """Reset progress bars for next epoch"""
        self.progress_bar.reset(self.train_bar)
        self.progress_bar.reset(self.val_bar)

    def __exit__(self, *args) -> None:
        """Clear per epoch bars and stop live display"""
        self.progress_bar.disable = True
        self.progress_bar.update(self.train_bar, visible=False)
        self.progress_bar.update(self.val_bar, visible=False)
        self.live.stop()

    def to_csv(self, savepath: str) -> None:
        df = {
            col.header: col._cells 
            for col in self.metrics_table.__dict__['columns']
        }
        DataFrame.from_dict(df).to_csv(savepath, index=False)
        print(f"Training logs saved @ `{savepath}`")
        



import numpy as np
import matplotlib.pyplot as plt
from typing import Tuple


__all__ = [
    # Functions
    'minmaxscaler',

    # Classes
    'OptmWithLRSWrapper',
]

# ============================= FUNCTIONS ==============================
def minmaxscaler(x: list, axis=0, min=0, max=1):
        x_std = (x - x.min(axis=axis)) / (x.max(axis=axis) - x.min(axis=axis))
        x_scaled  = x_std * (max - min) + min
        return x_scaled

def jais_lr_scheduler_values(num_epochs: int,
                             num_iters_per_epochs: int,
                             warmup_pct: float = 0.20,
                             lr_range: Tuple[float, float] = (0.01, 1e-6),
                             verbose: bool = False) -> list:
    # Number of iterations / batches in one epoch
    max_lr, min_lr = lr_range
    num_iters = num_iters_per_epochs
    n_steps = num_epochs * num_iters  # Total steps 
    n_warmup_steps = int(n_steps * warmup_pct)  # Number of warmup steps
    n_rem_steps = n_steps - n_warmup_steps  # Number of StepLR steps
    # Warmup lr values list
    warmup_lr = minmaxscaler(
        np.sqrt(np.linspace(min_lr, max_lr, num=n_warmup_steps)),
        min=min_lr, max=max_lr
    )
    n_steplr_epochs = int(n_rem_steps / num_iters) # Number of epochs for StepLR 
    # Remaining steps after warmup and StepLR
    n_rem_steps = n_steps - (n_warmup_steps + (n_steplr_epochs * num_iters))
    # StepLR values per epoch
    if n_rem_steps != 0: # Keep `min_lr` away for further adjustment
        steplr_epochs_lr = np.linspace(max_lr, min_lr * 10, num=n_steplr_epochs)
    else:
        # Do adjustments till `min_lr`
        steplr_epochs_lr = np.linspace(max_lr, min_lr, num=n_steplr_epochs)
    steplr_lr = np.repeat(steplr_epochs_lr, num_iters)  # StepLR values per 
    if n_rem_steps != 0:
        rem_lr = list(np.linspace(min_lr*10, min_lr, num=n_rem_steps))
        total_lr = list(warmup_lr) + list(steplr_lr) + rem_lr
    else:
        total_lr = list(warmup_lr) + list(steplr_lr)
    assert len(total_lr) == n_steps, f"total_lr = {len(total_lr)} but n_steps = {n_steps}."
    if verbose:
        plt.figure(figsize=(10,5))
        plt.title("JAIS Custom Learning Rate Scheduler")
        plt.plot(range(len(total_lr)), total_lr)
    return total_lr


# ============================= CLASSES ==============================
class OptmWithLRSWrapper:
    def __init__(self,
                 optm_fn,
                 num_epochs: int,
                 num_iters_per_epochs: int,
                 warmup_pct: float = 0.20,
                 lr_range: Tuple[float, float] = (0.01, 1e-6)) -> None:
        """Learning Rate Scheduler Wrapper"""
        self.optm_fn = optm_fn
        self.step_num = 0
        self.lr_list = jais_lr_scheduler_values(
            num_epochs=num_epochs,
            num_iters_per_epochs=num_iters_per_epochs,
            warmup_pct=warmup_pct,
            lr_range=lr_range)

    def step(self) -> None:
        "Step with the inner optimizer"
        self.update_lr()
        self.optm_fn.step()

    def zero_grad(self):
        "Zero out the gradients with the inner optimizer"
        self.optm_fn.zero_grad()
    
    def update_lr(self) -> None:
        try:
            self.current_lr = self.lr_list[self.step_num]
        except IndexError:
            # When last batch has less samples than others
            self.current_lr = self.lr_list[-1]
        for param_group in self.optm_fn.param_groups:
            param_group['lr'] = self.current_lr
        self.step_num += 1

    def get_lr(self) -> float:
        return self.current_lr


import torch
import torchmetrics
import torchvision
import torchvision.transforms as transforms
import torch.optim as optim
from torchsummary import summary
from jais.utils import *
from jais.train.trainer import Trainer
from jais.visualize.data import show_batch
from jais.models.example import Net
from jais.train.lr_scheduler import OptmWithLRSWrapper

CNF, LOG = load_default_configs()
LOG.info("Running CIFAR10 experiment...")

device, cuda_ids = get_device()
LOG.info(f"device = {device}")

NUM_EPOCHS = 20

train_tfms = transforms.Compose([
    transforms.AutoAugment(),
    transforms.ToTensor(),
    transforms.Normalize(CNF.data.mean, CNF.data.std)
])
val_tfms = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize(CNF.data.mean, CNF.data.std)
])

# DATASET
classes = ('plane', 'car', 'bird', 'cat', 'deer', 
           'dog', 'frog', 'horse', 'ship', 'truck')
n_classes = len(classes)

trainset = torchvision.datasets.CIFAR10(root=CNF.paths.data_dir, train=True,
                                        download=True, transform=train_tfms)
trainloader = torch.utils.data.DataLoader(trainset, batch_size=CNF.bs,
                                          shuffle=True, num_workers=2)
testset = torchvision.datasets.CIFAR10(root=CNF.paths.data_dir, train=False,
                                       download=True, transform=val_tfms)
testloader = torch.utils.data.DataLoader(testset, batch_size=CNF.bs,
                                         shuffle=False, num_workers=2)

# get some random training images
images, labels = next(iter(trainloader))
LOG.debug(f"IMAGES = {images.shape}")
LOG.debug(f"LABELS = {labels}")

# show_batch(images.permute(0,2,3,1), labels)

dls = {'train': trainloader, 'val': testloader}
NUM_BATCHES = len(trainset) // CNF.bs
# MODEL
LOG.debug("Loading model...")
net = Net()
# LOG.debug("Loading model summary...")
# summary(net, input_size=(3,32,32))

# LOSS
loss_fn = torch.nn.CrossEntropyLoss()

# OPTIMIZER AND LR SCHEDULER
optimizer = optim.SGD(net.parameters(), 
                      lr=CNF.train.lr, 
                      momentum=CNF.train.mom
                      )
optimizer = OptmWithLRSWrapper(optimizer, 
                               num_epochs=NUM_EPOCHS,
                               num_iters_per_epochs=NUM_BATCHES,
                               lr_range=(0.01, 1e-5))

metrics = torchmetrics.MetricCollection([
    torchmetrics.Accuracy(num_classes=n_classes, average='macro'),
    torchmetrics.Precision(num_classes=n_classes, average='macro'),
    torchmetrics.Recall(num_classes=n_classes, average='macro'),
    torchmetrics.F1(num_classes=n_classes, average='macro'),
])


def train_one_batch(self, batch):
    # DATA
    inputs, targets = batch
    inputs, targets = inputs.to(self.device), targets.to(self.device)
    # PREDICT
    outputs = self.net(inputs)
    # LOSS
    loss = self.loss_fn(outputs, targets)
    # BACKWARD
    self.optm_fn.zero_grad()
    loss.backward()
    self.optm_fn.step()
    return loss, outputs, targets

def val_one_batch(self, batch):
    # DATA
    inputs, targets = batch
    inputs, targets = inputs.to(self.device), targets.to(self.device)
    # PREDICT
    outputs = self.net(inputs)
    # LOSS
    loss = self.loss_fn(outputs, targets)
    return loss, outputs, targets


Trainer.train_one_batch = train_one_batch
Trainer.val_one_batch = val_one_batch
trainer = Trainer(dls=dls, net=net,
                  loss_fn=loss_fn,
                  optm_fn=optimizer,
                  device=device,
                  metrics=metrics, logger_name='wb')
trainer.train(20)



import os
import math
import time
import torch
from typing import Dict, Optional, Any
from collections import OrderedDict
from torchmetrics import Metric
from jais.train.log import TrainingLogger, RichProgress
from jais.utils import ROOT_DIR, load_default_configs, manage_log_files


__all__ = ['Trainer']

CNF, LOG = load_default_configs()


class Trainer:
    def __init__(self, 
                 dls: Dict[str, torch.utils.data.DataLoader], 
                 net: torch.nn.Module, 
                 loss_fn: torch.nn.Module, 
                 optm_fn: torch.nn.Module, 
                 device: torch.device, 
                 metrics: Optional[Dict[str, Metric]] = None, 
                 logger_name: Optional[str] = None) -> None:
        """Training and validation handler class

        Args:
            dls: dict of dataloaders
            net: Network / model to train and validate
            loss_fn: criterion instance
            optm_fn: optimizer instance
            device: device to train on.
            metrics: dict of metrics such as accuracy, precision, recall, etc.
            logger_name: Name of the logger to use
        """
        self.dls, self.net = dls, net.to(device)
        self.loss_fn, self.optm_fn = loss_fn ,optm_fn
        self.device, self.metrics = device, metrics
        self.logger = TrainingLogger(logger_name) if logger_name else None

        # Put weights tensor on same device
        if hasattr(self.loss_fn, 'weight') and \
            (self.loss_fn.weight is not None):
            self.loss_fn.weight = self.loss_fn.weight.to(self.device)

    def train_one_batch(self, batch):
        # return loss, outputs, targets
        """
        Either inherit this class to implement new routine
            OR
            def my_train_one_batch_func(self, batch):
                # Routine here
            Trainer.train_one_batch = my_train_one_batch_func
        """
        # DATA
        inputs  = batch['inputs' ].to(self.device)
        targets = batch['targets'].to(self.device)
        # PREDICT
        outputs = self.net(inputs)
        # LOSS
        loss = self.loss_fn(outputs, targets)
        # BACKWARD
        self.optm_fn.zero_grad()
        loss.backward()
        self.optm_fn.step()
        return loss, outputs, targets

    def val_one_batch(self, batch):
        """
        Either inherit this class to implement new routine
            OR
            def my_val_one_batch_func(self, batch):
                # Routine here
            Trainer.val_one_batch = my_val_one_batch_func
        """
        # DATA
        inputs  = batch['inputs' ].to(self.device)
        targets = batch['targets'].to(self.device)
        # PREDICT
        outputs = self.net(inputs)
        # LOSS
        loss = self.loss_fn(outputs, targets)
        return loss, outputs, targets

    def train_and_val_one_epoch(self, 
                                epoch_num: int,
                                rp: RichProgress = None,
                                max_train_iters: int = None,
                                max_val_iters: int = None) -> Dict[str, Any]:
        """Train and validate one epoch"""
        # Put network in training mode and device
        self.net.train()
        # Train one epoch
        epoch_loss, epoch_outputs, epoch_targets = [], [], []
        for i, batch in enumerate(self.dls['train'], start=1):
            loss, outputs, targets = self.train_one_batch(batch)
            epoch_loss.append(loss)
            epoch_outputs.append(outputs)
            epoch_targets.append(targets)
            rp.update_train_bar()
            if i >= max_train_iters:
                break
        # Compute training metrics
        avg_epoch_loss = torch.tensor(epoch_loss).mean().item()
        # epoch_outputs = torch.cat(epoch_outputs, dim=0).to(self.device)
        # epoch_targets = torch.cat(epoch_targets, dim=0).to(self.device)

        # Validate
        with torch.no_grad():
            self.net.eval()
            val_loss, val_outputs, val_targets = [], [], []
            for i, batch in enumerate(self.dls['val'], start=1):
                loss, outputs, targets = self.val_one_batch(batch)
                val_loss.append(loss)
                val_outputs.append(outputs)
                val_targets.append(targets)
                rp.update_val_bar()
                if i >= max_val_iters:
                    break
            # Compute validation metrics
            avg_val_loss = torch.tensor(val_loss).mean().item()
            val_outputs = torch.cat(val_outputs, dim=0).cpu()
            val_targets = torch.cat(val_targets, dim=0).cpu()
            val_metrics_dict = self.metrics(val_outputs, val_targets)
            val_metrics_dict = {
                k: v.item() for k, v in val_metrics_dict.items()
            }

        # Update metrics table and logger
        epoch_logs = OrderedDict({
            'epoch_num': epoch_num, 
            'train_loss': avg_epoch_loss,
            'val_loss': avg_val_loss,
        })
        epoch_logs.update(val_metrics_dict)
        return epoch_logs


    def train(self,
              n_epochs: int,
              max_train_iters: int = None,
              max_val_iters: int = None,
              optm_fn = None):
        if optm_fn:  # change optimizer
            self.optm_fn = optm_fn
        self.table_columns = ['Epoch', 'TrainLoss']
        if max_train_iters is None:
            max_train_iters = math.ceil(
                len(self.dls['train'].dataset) / self.dls['train'].batch_size
            )
        if (max_val_iters is None) and ('val' in self.dls.keys()):
            max_val_iters = math.ceil(
                len(self.dls['val'].dataset) / self.dls['val'].batch_size
            )
        if 'val' in self.dls.keys():
            self.table_columns.append('ValLoss')
        # Add metrics names as table columns
        self.table_columns += [
            f"Val{k.capitalize()}" for k in self.metrics.keys()
        ]

        with RichProgress(columns=self.table_columns, 
                          n_epochs=n_epochs, 
                          n_train_iters=max_train_iters,
                          n_val_iters=max_val_iters) as rp:
            for epoch_num in range(1, n_epochs + 1):
                
                epoch_logs = self.train_and_val_one_epoch(
                    epoch_num=epoch_num,
                    rp=rp,
                    max_train_iters=max_train_iters, 
                    max_val_iters=max_val_iters
                )
                if self.logger:
                    self.logger.log(epoch_logs)
                rp.update_epoch(list(epoch_logs.values()))

        # Save logs
        savepath = f"{CNF.paths.logs_dir}/{CNF.log.training_log_filename}@{time.time()}.csv"
        manage_log_files(CNF, '.csv')
        rp.to_csv(savepath)


from setuptools import setup, find_packages
from jais.utils import load_default_configs

CNF = load_default_configs()

VERSION = '0.0.1.1'

f = open('README.md', 'r')
LONG_DESCRIPTION = f.read()
f.close()


    
def read_requirements(path):
    with open(path) as f:
        req_list = f.read().split('\n')
    return [l.strip() for l in req_list]

setup(
    name="jais",
    version=VERSION,
    description='Just Artificial Inteliigence Snippets (JAIS)',
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    author='Jitender Singh Virk',
    author_email='krivsj@gmail.com',
    url='https://virksaab.github.io',
    # license='unlicensed',
    packages=find_packages(exclude=['ez_setup', 'tests*', ".github"]),
    # package_data={'myproject': ['templates/*']}, ?? WHAT IS THIS FOR?
    include_package_data=True,
    install_requires=read_requirements("requirements.txt"),
    entry_points=f"""
        [console_scripts]
        jais=jais.__main__:cli
    """,
)



"""This module contains some useful functions which can be used throughout 
the project.
"""

from cgitb import handler
import os
import time
import click
import logging
from rich import print
from typing import Any, Callable, Tuple, Union, Dict, Optional
from ruamel.yaml import YAML, yaml_object
from itertools import cycle
from shutil import get_terminal_size
from pathlib import Path
from easydict import EasyDict as edict
from collections import OrderedDict


__all__ = [
    # Configuration handlers
    'load_config', 'save_config',

    # Functions
    'get_recent_githash', 'get_logger', 'get_device', 'install_rich',
    'load_default_configs',

    # Classes
    'SpinCursor',

    # Decorators
    'show_exec_time',

    # Variables
    'ROOT_DIR'
]

# ======================= CONFIGURATION HANDLERs ===========================

# * When using the decorator, which takes the YAML() instance as a parameter,
# * the yaml = YAML() line needs to be moved up in the file -- *yaml docs*
yaml = YAML(typ='safe', pure=True)
yaml.default_flow_style = False
yaml.indent(mapping=2, sequence=4, offset=2)
ROOT_DIR = Path(__file__).parent # Root for the JAIS package, not project.


@yaml_object(yaml)
class JoinPath:
    """Custom tag `!join` loader class to join strings for yaml file."""

    yaml_tag = u'!joinpath'

    def __init__(self, joined_string):
        self.joined_string = joined_string

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag,
                                            u'{.joined_string}'.format(node))

    @classmethod
    def from_yaml(cls, constructor, node):
        seq = constructor.construct_sequence(node)
        fullpath = Path('/'.join([str(i) for i in seq])).resolve()
        if len(str(fullpath.name).split(".")) == 1:  # This is a directory
            fullpath.mkdir(parents=True, exist_ok=True)
            # Create a empty .gitkeep file to keep the empty folder structure
            # in git repo
            if len(list(fullpath.glob("**/*"))) == 0:
                (fullpath/".gitkeep").touch(mode=0o666, exist_ok=True)
        return str(fullpath)


@yaml_object(yaml)
class RootDirSetter:
    """Custom tag `!rootdir` loader class for yaml file."""

    yaml_tag = u'!rootdir'

    def __init__(self, path):
        self.path = path

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag,
                                            u'{.path}'.format(node))

    @classmethod
    def from_yaml(cls, constructor, node):
        return str(ROOT_DIR)


@yaml_object(yaml)
class HomeDirSetter:
    """Custom tag `!rootdir` loader class for yaml file."""

    yaml_tag = u'!homedir'

    def __init__(self, path):
        self.path = path

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag,
                                            u'{.path}'.format(node))

    @classmethod
    def from_yaml(cls, constructor, node):
        return str(Path.home())


def load_config(path: Union[None, str, Path] = None, pure: bool = False) -> dict:
    """config.yaml file loader.

    This function converts the config.yaml file to `dict` object.

    Args:
        path: config.yaml filepath
        pure: If True, just load the .yaml without converting to EasyDict
            and exclude extra info.

    Returns:
        `dict` object containing configuration parameters.

    Example:
        .. code-block:: python

            config = load_config("../config.yaml")
            print(config["project_name"])
    """
    if path is None:
        path = os.getenv('JAIS_CNF_PATH')
        if path is None:
            _msg = "`[yellow]JAIS_CNF_PATH[/yellow]` environment variable "
            _msg += "containing path to the configuration file is not set. "
            _msg += "Loaded default configuration file from "
            _msg += "[yellow]jais/configs/default.yaml[/yellow]. "
            _msg += "Set `[yellow]JAIS_CNF_PATH[/yellow]` to your configuration file to load "
            _msg += "different settings"
            print(_msg)
            path = ROOT_DIR/"configs/default.yaml"

    path = str(Path(path).absolute().resolve())
    # * Load config file
    with open(path) as file:
        config = yaml.load(file)

    if pure == False:  # Add extra features
        # Convert dict to easydict
        config = edict(config)
        # Save the config filepath itself for further convenience
        config.original_config_filepath = path
    return config


def save_config(config_dict: Dict, path: Union[str, Path] = None,
                saveas_ordered: bool = True, force_overwrite: bool = False,
                file_extension: str = 'yaml') -> None:
    """save `dict` config parameters to `.yaml` file

    Args:
        config_dict: parameters to save.
        path: path/to/save/auto_config.yaml.
        saveas_ordered: save as collections.OrderedDict on Python 3 and 
            `!!omap` is generated for these types.
        file_extension: default `.yaml`

    Returns: 
        nothing

    Example:
        .. code-block:: python

            config = {"Example": 10}
            save_config(config, "../auto_config.yaml")
    """

    if path is None:  # set default path if not given
        path = config_dict.paths.output_dir
        path = f"{path}/modified_config.{file_extension}"
    else:
        path = str(Path(path).absolute().resolve())
        # Check if the path given is the original config's path
        if (config_dict.original_config_filepath == path) and \
           (not force_overwrite):
            msg = f"""
            Error while saving config file @ {path}.
            Cannot overwrite the original config file
            Choose different save location.
            """
            raise ValueError(msg)

    # converting easydict format to default dict because
    # YAML does not processes EasyDict format well.
    cleaned_dict = {
        k: edict_to_dict_converter(v)
        for k, v in config_dict.items()
    }
    # print("cleaned_dict =", cleaned_dict)

    # Fix order
    if saveas_ordered:
        cleaned_dict = OrderedDict(cleaned_dict)

    # Save the file to given location
    with open(path, 'w') as file:
        yaml.dump(cleaned_dict, file)

    print(f"config saved @ {path}")


def edict_to_dict_converter(x: Union[Dict, Any]) -> Union[Dict, Any]:
    """Recursive function to convert given dictionary's datatype
    from edict to dict.

    Args:
        x (dict or other): nested dictionary

    Returns: x (same x but default dict type)
    """
    if not isinstance(x, dict):
        return x

    # Recursion for nested dicts
    ndict = {}
    for k, v in x.items():
        v1 = edict_to_dict_converter(v)
        ndict[k] = v1
    return ndict


# ============================== CLASSES ==============================
class SpinCursor:
    """A waiting animation when program is being executed. 

        `reference source (stackoverflow) <https://stackoverflow.com/questions/22029562/python-how-to-make-simple-animated-loading-while-process-is-running>`_ 

        Args:
            desc : The loader's description. Defaults to "Loading...".
            end : Final print. Defaults to "Done!".
            cursor_type : Set the animation type. Choose one out of
                'bar', 'spin', or 'django'.
            timeout : Sleep time between prints. Defaults to 0.1.

        Example:
            Using *with* context:

            .. code-block:: python

                with SpinCursor("Running...", end=f"done!!"):
                    subprocess.run(['ls', '-l'])
                    time.sleep(10)

            Using normal code:

            .. code-block:: python

                cursor = SpinCursor("Running...", end=f"done!!")
                cursor.start()
                subprocess.run(['ls', '-l'])
                time.sleep(10)
                cursor.stop()

        Returns: 
            Nothing
        """

    def __init__(self, desc: str = "Loading...", end: str = "Done!", cursor_type: str = "bar", timeout: float = 0.1) -> None:
        from threading import Thread
        self.desc = desc
        self.end = end
        self.timeout = timeout

        self._thread = Thread(target=self._animate, daemon=True)

        if cursor_type == 'bar':
            self.steps = [
                "[=     ]",
                "[ =    ]",
                "[  =   ]",
                "[   =  ]",
                "[    = ]",
                "[     =]",
                "[    = ]",
                "[   =  ]",
                "[  =   ]",
                "[ =    ]",
            ]
        elif cursor_type == 'spin':
            self.steps = ['|', '/', '-', '\\']
        elif cursor_type == 'django':
            self.steps = ["⢿", "⣻", "⣽", "⣾", "⣷", "⣯", "⣟", "⡿"]
        else:
            raise NotImplementedError("choose one [`spin`, `bar`, `django`].")

        self.done = False

    def start(self) -> object:
        """Start the animation. See example above."""
        self._thread.start()
        return self

    def _animate(self) -> None:
        for c in cycle(self.steps):
            if self.done:
                break
            print(click.style(f"\r{self.desc} {c}",
                  fg='yellow'), flush=True, end="")
            time.sleep(self.timeout)

    def __enter__(self) -> None:
        self.start()

    def stop(self) -> None:
        """Stop animation. See example above."""
        self.done = True
        cols = get_terminal_size((80, 20)).columns
        print("\r" + " " * cols, end="", flush=True)
        print(click.style(f"\r{self.end}", fg='green'), flush=True)

    def __exit__(self, exc_type, exc_value, tb):
        # handle exceptions with those variables ^
        self.stop()


# ============================== DECORATORS ==============================
def show_exec_time(func: Callable) -> Any:
    """Display the execution time of a function.

        This decorator is suited to large programs that takes
            more than a second to run.

        Example:

            .. code-block:: python

                @show_exec_time
                def take_a_break(timeout=10):
                    time.sleep(timeout)

                >>> take_a_break()
                >>> >> Completed in 00h:00m:10s <<
    """
    def wrapper(*args, **kwargs):
        start_time = time.time()

        # Run the given function
        results = func(*args, **kwargs)

        end_time = time.time()

        hrs = (end_time - start_time) // 3600
        rem = (end_time - start_time) % 3600
        mins = rem // 60
        secs = rem % 60

        hrs = str(round(hrs)).zfill(2)
        mins = str(round(mins)).zfill(2)
        secs = str(round(secs)).zfill(2)

        print(f"\n>> Completed in {hrs}h:{mins}m:{secs}s <<\n")

        return results
    return wrapper


# ============================= FUNCTIONS ==============================
def get_recent_githash():
    """Get the recent commit git hash"""
    import subprocess
    proc = subprocess.Popen(["git", "rev-parse", "HEAD"],
                            stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    return proc.stdout.read().strip().decode('ascii')


def get_logger(CNF: Optional[Union[dict, edict]] = None,
               log_filepath: Optional[Union[str, Path]] = None,
               logs_conf_filepath: Optional[Union[str, Path]] = None,
               level: Optional[str] = None,
               rich_logger: bool = True,
               colored_logger: bool = False) -> logging.Logger:
    """Create a colored logger with two handlers (stream and file logging)
        from logs.conf

    Args:
        CNF: loaded configurations from `configs/default.yaml`. 
            Or pass a dict with the following:
                CNF = EasyDict({
                    'paths': {'logs_dir': 'path/to/logs/folder'},
                    'log': {'name': 'JAISLogger', 'filename': 'jais_logs'}
                })
        log_filepath: File path to save logs to a file. Otherwise default file
            will be used using configs/default.yaml:log information
        logs_conf_filepath: Logs configuration filepath. 
            [Default is configs/logs.conf]
        level: override JAISLogger.StreamHandler logging level: 
            ['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'].
            Default is loaded from the file.
        rich_logger: Use `RichHandler` logger for colored logs
        colored_logger: Use `coloredlogs` logger for colored logs.
            Mutually exclusive to rich_logger.
    Returns:
        Colored logger instance with settings loaded from `configs/logs.conf`
    """
    if (rich_logger == True) and (colored_logger == True):
        _errmsg = "Both `rich_logger` and `colored_logger` cannot be\
             True at the same time."
        raise ValueError(_errmsg)

    import configparser  # To read logs.conf file
    import logging
    import logging.config

    if CNF is None:
        CNF = load_config(ROOT_DIR/'configs/default.yaml')
    if logs_conf_filepath is None:
        logs_conf_filepath = ROOT_DIR/"configs/logs.conf"
    logs_conf = configparser.RawConfigParser()
    logs_conf.read(logs_conf_filepath)
    default_level = logs_conf['logger_JAISLogger']['level']
    if log_filepath is None:
        log_filepath = f"{CNF.paths.logs_dir}/{CNF.log.filename}@{time.time()}.log"
    logging.config.fileConfig(logs_conf_filepath,
                              defaults={'logfilename': log_filepath},
                              disable_existing_loggers=True)
    # Create logger
    logger = logging.getLogger(CNF.log.name)
    # Override logging level
    if level:
        logger.setLevel(level)
        default_level = level

    if rich_logger:
        # Existing StreamHandler needs to be removed, if present.
        existing_handlers = [
            handler for handler in logger.handlers
            if isinstance(handler, logging.FileHandler)
        ]
        logger.handlers.clear()
        logger.handlers = existing_handlers
        # Add RichHandler
        from rich.logging import RichHandler
        ch = RichHandler(level=default_level,
                         show_time=False,
                         rich_tracebacks=True)
        ch.setLevel(default_level)
        formatter = logging.Formatter('%(message)s')
        ch.setFormatter(formatter)
        logger.addHandler(ch)
    elif colored_logger:
        import coloredlogs
        # StreamLogger format settings for coloredlogs
        stream_fmt = '[%(programname)s - %(levelname)s] %(message)s'
        # Set colored logs
        coloredlogs.install(level=default_level, fmt=stream_fmt, logger=logger)

    manage_log_files(CNF)
    return logger


def manage_log_files(CNF: Union[dict, edict], file_ext: str = '.log'):
    """Log files rotation handler"""
    CNF = edict(CNF)
    # Get log files paths
    log_filespaths = list(Path(CNF.paths.logs_dir).glob(f"*{file_ext}"))
    # Function to split the timestamp from filepath
    def get_time_from_filename(x): return float(x.stem.split('@')[-1])
    # Sort timestamps
    timestamps = sorted(
        list(map(get_time_from_filename, log_filespaths)),
        reverse=True)
    # Keep only n recent files
    timestamps = timestamps[: CNF.log.keep_n_recent_logs]
    # Remove old files
    for fp in log_filespaths:
        if get_time_from_filename(fp) not in timestamps:
            os.remove(fp)


def get_device():
    """Get torch device instance and available GPU IDs"""
    from torch.cuda import device_count
    from torch import device
    cuda_ids = [0] if device_count() == 1 else list(range(device_count()))
    return device(f"cuda:{cuda_ids[0]}"), cuda_ids


def install_rich(verbose: bool = False):
    """Enable Rich to override Python """
    from rich import pretty, traceback
    import click
    pretty.install()
    # If you are working with a framework (click, django etc),
    # you may only be interested in seeing the code from your own
    # application within the traceback. You can exclude framework
    # code by setting the suppress argument on Traceback, install,
    # and Console.print_exception, which should be a list of modules
    # or str paths.
    traceback.install(show_locals=False, suppress=[click])
    if verbose:
        print("[cyan]Rich set to override print and tracebacks.")


def load_default_configs() -> Tuple[edict, logging.Logger]:
    """Load the default configuration settings and logger"""
    install_rich(verbose=False)
    # Load configurations
    CNF_PATH = os.getenv('JAIS_CNF_PATH')
    if CNF_PATH is None:
        CNF_PATH = f"{ROOT_DIR}/configs/default.yaml"
        os.environ['JAIS_CNF_PATH'] = CNF_PATH
    CNF = load_config(CNF_PATH)

    # Set logger settings
    LOG_PATH = os.getenv('JAIS_LOG_PATH')
    if LOG_PATH is None:
        LOG_PATH = f"{CNF.paths.logs_dir}/{CNF.log.filename}@{time.time()}.log"
        os.environ['JAIS_LOG_PATH'] = LOG_PATH

    # Set logger config filepath
    LOG_CNF_PATH = os.getenv('JAIS_LOG_CNF_PATH')
    if LOG_CNF_PATH is None:
        LOG_CNF_PATH = f"{ROOT_DIR}/configs/logs.conf"
        os.environ['JAIS_LOG_CNF_PATH'] = LOG_CNF_PATH
    LOG = get_logger(CNF, LOG_PATH, LOG_CNF_PATH)

    return CNF, LOG
